/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2015 Satellite Multimedia Group (TLCSat)
 * http://www.tlcsat.uniroma2.it
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * Authors: A. Abdelsalam, M. Luglio, C. Roseti and F. Zampognaro, <{abdel.salam, roseti, zampognaro}@ing.uniroma2.it>
 *
 * *** brief TcpWave implementation *****
 * TCP Wave is a new transport protocol designed to provide a reliable transport services over the current and future networks.
 * This goal is achieved through an innovative burst-based transmission paradigm, as an alternative of the traditional TCP window based approach.
 * TCP Wave design principles have been tailored to assure an optimal performance under time-varying network topologies.
 * For detailed design principals and validation, please refer to:
 * To follow the protocol implementation and how the burst transmission is performed, start from the function
 * TcpWave::SendPendingData.
 */

#define NS_LOG_APPEND_CONTEXT \
		if (m_node) \
		{ std::clog << Simulator::Now ().GetSeconds () << \
			" [node " << m_node->GetId () << "] "; }

#include "tcp-wave.h"
#include "ns3/node.h"
#include "ns3/log.h"
#include "ns3/abort.h"

using namespace ns3;
NS_LOG_COMPONENT_DEFINE("TcpWave");
NS_OBJECT_ENSURE_REGISTERED(TcpWave);

#define LAMBDA 2
#define STAB 1
#define ALPHA 0.9
#define BURST_MIN (m_segmentSize * 3)

TypeId
TcpWave::GetTypeId(void)
{
	static TypeId tid = TypeId("ns3::TcpWave")
					.SetGroupName ("Internet")
					.SetParent<TcpSocketBase>()
					.AddConstructor<TcpWave>()
					.AddAttribute("TxTime","Default transmission timer", TimeValue(MilliSeconds(500)),
							MakeTimeAccessor(&TcpWave::m_initialTxTime),
							MakeTimeChecker()).AddAttribute("B", "Congestion Threshold", TimeValue(MilliSeconds(200)),
									MakeTimeAccessor(&TcpWave::m_congThresold),
									MakeTimeChecker());
	return tid;
}

TcpWave::TcpWave() :
TcpSocketBase(),
m_txTimer(Timer::CANCEL_ON_DESTROY),
m_trainReceived(STAB - 1),
m_bytesAcked(0),
m_maxReTx(0),
m_minRtt(Time::Max()),
m_maxRtt(Time::Min()),
m_avgRtt(0),
m_minAbsRtt(Time::Max()),
m_sf(STAB),
m_minTrainDispersion(Time::FromInteger(100, Time::MS)),
m_firstAck(Time::FromInteger(0, Time::MS)),
m_rtoTxThreshold(0),
m_bytesRetransmitted(0),
m_bytesTransmitted(0),
m_recovery(false),
m_txmode(0),
m_ACCEProcessing(false)
{
	SetTcpNoDelay(false);
	isTimerEnabled = false;
	m_cWnd = m_bWnd = m_initialBWnd = 0;
}

void TcpWave::SetInitialSSThresh(uint32_t threshold) {
	(void) threshold;
	NS_LOG_WARN("TcpWave does not perform Slow-Start Phase");
}

uint32_t TcpWave::GetInitialSSThresh(void) const {
	NS_LOG_WARN("TcpWave does not perform Slow-Start Phase");
	return 0;
}

void TcpWave::SetInitialCwnd(uint32_t cwnd) {
	NS_LOG_FUNCTION (this);
	NS_ABORT_MSG_UNLESS(m_state == CLOSED,
			"TcpWave::SetInitialCwnd() cannot change initial cwnd after connection started.");
	m_initialBWnd = cwnd;
}

uint32_t TcpWave::GetInitialCwnd(void) const {
	NS_LOG_FUNCTION (this);
	return m_initialBWnd;
}
int
TcpWave::Listen (void)
{
	NS_LOG_FUNCTION (this);
	InitializeCwnd();
	return TcpSocketBase::Listen();
}

int TcpWave::Connect(const Address &address) {
	NS_LOG_FUNCTION(this);
	int res = TcpSocketBase::Connect(address);
	InitializeCwnd();
	return res;
}

void
TcpWave::CancelAllTimers (void)
{
	NS_LOG_FUNCTION (this);
	m_txTimer.Cancel();
	return TcpSocketBase::CancelAllTimers ();
}
void
TcpWave::InitializeCwnd()
{
	NS_LOG_FUNCTION(this);
	m_initialBWnd = m_initialBWnd * m_segmentSize;
	m_bWnd = m_initialBWnd;
	m_txTime = m_initialTxTime;
}

void
TcpWave::StartTxTimer()
{
	NS_LOG_FUNCTION(this);
	NS_ASSERT(!m_txTimer.IsRunning());
	NS_ASSERT(m_txTime.Get().GetSeconds() != 0.0);
	m_txTimer.SetFunction(&TcpWave::CallSendPendingData, this);
	if (!isTimerEnabled) {
		m_txTimer.SetDelay(Time::FromInteger(1, Time::MS));
		isTimerEnabled = true;
	} 	else {
		m_txTimer.SetDelay(m_txTime);
	}
	m_txTimer.Schedule();
}

uint32_t
TcpWave::Window()
{
	NS_LOG_FUNCTION(this);
	if (!isTimerEnabled)
	{
		StartTxTimer();
	}
	return std::min (m_rWnd.Get (), m_cWnd);
}

uint32_t
TcpWave::AvailableWindow()
{
	NS_LOG_FUNCTION(this);
	uint32_t availableWindow = TcpSocketBase::AvailableWindow();
	return availableWindow;
}

void
TcpWave::CallSendPendingData()
{
	NS_LOG_FUNCTION(this);

	/* Re-schedule the TXTimer if the connection is not yet established */
	if (m_state != ESTABLISHED)
	{
		StartTxTimer();
		return ;
	}
	uint32_t cWnd = m_bWnd; // local value to calculate the right amount of bytes to be transmitted //

	if (m_bytesTransmitted != 0 && m_historyBWnd.size() > 0)
	{
		NS_ASSERT(m_historyBWnd.size() > 0);
		/* Update structure of Burst Sizes with the effective amount of packets sent in the previous SendPendingData
		 * This is done updating the last element, before appending the new one. */
		m_historyBWnd[m_historyBWnd.size() - 1] = m_bytesTransmitted;
	}

	m_bytesTransmitted = 0; // once updated to the history, reset the counter.

	/* taking care of the previously re-transmitted byte
	 * reduce the upcoming burst size, to preserve the right overall amount of bytes on the network.
	 * */
	if (cWnd > BURST_MIN && m_bytesRetransmitted > 0)
	{
		cWnd -= m_bytesRetransmitted;

	}
	m_cWnd = UnAckDataCount() + cWnd; // adding the amount of bytes to be transmitted to the UnAckDataCount().
	m_cWnd = std::min(m_txBuffer->Size(), m_cWnd); // avoid overcoming TxBuffer size.
	cWnd = m_cWnd - UnAckDataCount(); //check for the flow control

	// to adjust the send / receive window size
	if (cWnd > m_rWnd.Get())
	{
		cWnd = m_rWnd.Get();
	}

	m_bWnd = cWnd;

	if (m_state != ESTABLISHED || m_bWnd == 0)
	{
		/*		1) as soon as data is over, i enter here and don't reschedule timer
		 * 		2) as soon as data is back, window() is called, to restore the timer (of which first
		 * 		call is done after 1 ms and the following ones after tx_timer)
		 **/
		isTimerEnabled = false;
		m_bWnd = m_initialBWnd;
		return ;
	}


	if (cWnd != 0 && m_recovery == false)
	{

		m_historyBWnd.push_back(cWnd);
	}

	m_bytesRetransmitted = 0;
	SendPendingData(true);

	//*DEBUG* printout the contents of BWNDHistory
	unsigned int i;
	for (i = 0; i < m_historyBWnd.size(); i++)
	{
		NS_LOG_DEBUG(
				"historyBW: "<< i <<" size: " << m_historyBWnd[i] << " Segments: " << m_historyBWnd[i] /m_segmentSize);
	}
}

bool
TcpWave::SendPendingData(bool withAck)
{
	NS_LOG_FUNCTION(this << withAck);
	if (m_cWnd > BytesInFlight() && !m_recovery
			&& m_cWnd - BytesInFlight() > std::min(m_rWnd.Get(), m_bWnd))
	{
		// Inflating m_cWnd at the present burst to allow transmission
		m_cWnd = BytesInFlight() + std::min(m_rWnd.Get(), m_bWnd);
	}

	if (m_recovery)
	{
		m_cWnd = UnAckDataCount() + std::min(m_rWnd.Get(), m_bWnd);
	}

	m_head = m_highTxMark.Get();
	SequenceNumber32 Retranshead = m_nextTxSequence.Get();
	uint32_t bytesTransmitted;

	/* Send the bytes through TcpSocketBase::SendPendingData */
	bool res = TcpSocketBase::SendPendingData(withAck);

	/* Update with the effective bytes transmitted
	 * (may be less than what calculated in the CallSendPendingData Function) */
	bytesTransmitted = m_highTxMark.Get() - m_head;

	if (m_recovery == false)
	{

		//*DEBUG* these are for outputs used for graphs & validation..
		if (m_txmode == true)
		{
			NS_LOG_DEBUG("TCPW BURST "<< " from: " << m_head << " to: " << m_highTxMark.Get() << " bytes " << " mode = tracking ");
		}
		if (m_txmode == false)
		{
			NS_LOG_DEBUG("TCPW BURST "<< " from: " << m_head << " to: " << m_highTxMark.Get() << " bytes " << " mode = adjustment ");
		}
		if (bytesTransmitted > 0 )
		{
			// Register the transmitted bytes sequence and the exact transmission time //
			Range32 rangeTransmitted(m_head, m_highTxMark.Get());
			PairOfRangeAndTime transmitTime(rangeTransmitted, Simulator::Now());
			m_historyTransmitTime.insert(m_historyTransmitTime.end(), transmitTime);
		}
	}

	// We are in the RTO, perform RTO recovery since we are still under the RTO captured threshold //
	// NO BWNDHistory updates here //
	if (m_rtoTxThreshold == m_highTxMark.Get() && m_recovery == true)
	{
		bytesTransmitted = 0;
		m_cWnd = UnAckDataCount();

		Range32 rangeTransmitted(Retranshead, m_nextTxSequence.Get());
		PairOfRangeAndTime transmitTime(rangeTransmitted, Simulator::Now());
		m_historyTransmitTime.insert(m_historyTransmitTime.end(),transmitTime);

		//*DEBUG* printout the RTO recovery bursts //
		NS_LOG_DEBUG(
				"TCPW BURST " << " from: " << Retranshead << " to: " << m_nextTxSequence.Get()<< " bytes "<<  " mode = in-rto recovery "" RTO threshold= "<< m_rtoTxThreshold<< " bytes ");
	}

	/* we finished with RTO and this is the first burst after the threshold,
	 * the bytes transmitted should be the difference between the hightx and the threshold...
	 */
	if (m_rtoTxThreshold < m_highTxMark.Get() && m_recovery == true) {
		bytesTransmitted = m_nextTxSequence.Get() - m_rtoTxThreshold;

		// special update to the BWNDhistory in the first call for this function after RTO recovery//
		if (bytesTransmitted > 0 && m_historyBWnd.size() >= 0)
		{
			m_historyBWnd.push_back(bytesTransmitted);
			Range32 rangeTransmitted(Retranshead, m_highTxMark.Get());
			PairOfRangeAndTime transmitTime(rangeTransmitted, Simulator::Now());
			m_historyTransmitTime.insert(m_historyTransmitTime.end(),transmitTime);

			// time to safely turn off the recovery mode
			m_recovery = false;

			NS_LOG_DEBUG(
					"TCPW BURST " << " from: " << Retranshead << " to: " << " bytes " << " mode = after-rto recovery " " RTO Threshold= " <<m_rtoTxThreshold << " bytes ");
		}
	}

	NS_LOG_DEBUG("TCPW TX-TIME= " << m_txTime.Get().GetMilliSeconds() << " ms "); // printout the effective TxTime value
	NS_LOG_DEBUG("TCPW TX-RATE: " << (bytesTransmitted*8/m_txTime.Get ().GetSeconds()) << " bit/s "); // printout the effective Tx Rate

	NS_ASSERT(bytesTransmitted <= m_bWnd);

	m_bytesTransmitted += bytesTransmitted;

	if (!m_txTimer.IsRunning() && m_state <= FIN_WAIT_1) {

		StartTxTimer();
	}
	return res;
}

MapOfTransmitTime::iterator
TcpWave::GetIteratorOf(SequenceNumber32 seq)
{
	NS_LOG_FUNCTION(this);
	MapOfTransmitTime::iterator it;

	for (it = m_historyTransmitTime.begin(); it != m_historyTransmitTime.end();
			++it) {
		Range32 range = it->first;

		if (range.first <= seq && range.second >= seq) {
			return it;
		}
	}

	return it;

}

void
TcpWave::GetHistoryBWndContent()
{
	unsigned int i;
	unsigned int size = 0;
	for (i = 0; i < m_historyBWnd.size(); i++)
	{
		size += m_historyBWnd[i];
		NS_LOG_DEBUG(
				"historyBW: "<< i <<" size: " << m_historyBWnd[i] << " segm: " << m_historyBWnd[i] /m_segmentSize);
	}

}

Time
TcpWave::GetTransmitTime(SequenceNumber32 seq)
{
	MapOfTransmitTime::iterator it = GetIteratorOf(seq);

	if (it != m_historyTransmitTime.end())
	{
		return it->second;
	}
	return Simulator::Now();
}

void
TcpWave::DiscardTimeUpTo(SequenceNumber32 seq)
{
	MapOfTransmitTime::iterator it;
	for (it = m_historyTransmitTime.begin(); it != m_historyTransmitTime.end(); ++it)
	{
		Range32 range = it->first;
		if (range.second <= seq)
		{
			if (it != m_historyTransmitTime.end()) {
				m_historyTransmitTime.erase(it);
			}
		}
		if (range.first >= seq)
		{
			return;
		}
	}
}

void
TcpWave::NewAck(SequenceNumber32 const& rAck)
{
	NS_LOG_FUNCTION(this << rAck);
	Time now = Simulator::Now();
	Time rtt;
	if (m_state == SYN_RCVD)
	{
		TcpSocketBase::NewAck(rAck);
		return;
	}

	uint32_t bytesAcked = rAck - m_txBuffer->HeadSequence();
	uint32_t fullSegmentAcked = bytesAcked / m_segmentSize;
	uint32_t partialSegmentAcked  = bytesAcked % m_segmentSize;

	NS_ASSERT(bytesAcked >= 0);

	if (bytesAcked <= m_historyBWnd.front() && m_historyBWnd.front() > m_segmentSize) {
		m_ACCEProcessing = true;
	}

	// calculate RTT once we receive a new ACK then identify the min, max, minAbs RTT values
	rtt = now - GetTransmitTime(rAck);

	if (fullSegmentAcked < 2)
	{
		if (m_minAbsRtt > rtt || m_minAbsRtt == 0 ) {
			m_minAbsRtt = rtt;
		}

		if (m_minRtt > rtt) {
			m_minRtt = rtt;
		}

		if (m_maxRtt < rtt) {
			m_maxRtt = rtt;
		}
	}

	// here we handle the cumulative ACKs received by spletting it to multiple single ACKs
	uint32_t i;
	for (i = 1; i <= fullSegmentAcked; ++i)
	{

		if ((fullSegmentAcked * i) > m_segmentSize)
		{
			//*DEBUG*//
			NS_LOG_DEBUG("TCPW ACK seq# "<< rAck << " type= cumulative");

		}
		else
		{
			//*DEBUG*//
			NS_LOG_DEBUG("TCPW ACK seq# "<< rAck<< " type= single");
		}

		NS_LOG_DEBUG("TCPW RTT = " << rtt.GetMilliSeconds() << " ms "); //**DEBUG* printout the actual RTT measured.

		// Call the ACCE function with every single ACK
		ACCE(m_segmentSize, rtt, now);
	}

	if (partialSegmentAcked > 0 )
	{
		ACCE(partialSegmentAcked, rtt, now);
	}

	//After processing ACKs through ACCE free the records from the m_historyTransmitTime structure up to rack seq.
	DiscardTimeUpTo(rAck);

	//Avoid Sliding Window that would let new unhandled packet out from SockBase.
	if (UnAckDataCount() < BytesInFlight())
	{
		m_cWnd = 0;
	}
	else
	{
		if (m_cWnd >= bytesAcked)
		{
			m_cWnd -= bytesAcked;
		}

		else
		{
			m_cWnd = UnAckDataCount();
		}
	}

	//	once the RTX during RTO has been done successfully and segments have been acknowledged, we reset the RTO threshold.
	if (m_txBuffer->HeadSequence() > m_rtoTxThreshold) {
		m_rtoTxThreshold = 0;
	}

	TcpSocketBase::NewAck(rAck);
}

void
TcpWave::ACCE(uint32_t bytesAcked, Time rtt, Time now)
{

	NS_LOG_FUNCTION(this << bytesAcked << rtt.GetMilliSeconds ());
	Time ackTrainDispersion, ackDispersion, deltaRtt;

	uint32_t bWnd = m_historyBWnd.front();

	if (bWnd == 0) // if m_historyBWnd is empty, exit ACCE
	{
		return;
	}
	// history has a bust of 1 segment or less, we simply delete the record then return
	if (bWnd <= m_segmentSize)
	{
		m_historyBWnd.pop_front();
		m_ACCEProcessing = false;

		return;
	}
	// find out the time ACCE called by the first ACK in the train
	if (m_firstAck.IsZero())
	{
		m_firstAck = now;
		m_rtt->Measurement(rtt);
	}

	// count the bytes ACKED.
	m_bytesAcked += bytesAcked;

	// when bytes acked counter is equal to the fist BWND history record then start ACCE process.
	// ACK received at this moment is the last ACK in the train.

	if (m_bytesAcked >= bWnd)
	{
		// First ACCE FUNCTION: Timers Calculation and Update
		if (m_ACCEProcessing == true)
		{
			// 1- Calculate ACK train disp based on the first and last ACKS received.
			if ((now - m_firstAck) < m_avgRtt && (now - m_firstAck) > 0.00)
			{
				ackTrainDispersion = now - m_firstAck;
			}
			else
			{

				ackTrainDispersion = m_minTrainDispersion; // ACK Train disp couldn't be calculated yet. i.e. we don't know AvgRTT at the first train receieved
			}

			if (m_minTrainDispersion > ackTrainDispersion
					&& ackTrainDispersion > 0 && bWnd == m_initialBWnd)
			{
				m_minTrainDispersion = ackTrainDispersion;
			}

			NS_ASSERT(ceil((double)bWnd / m_segmentSize) - 1);

			// 2- Calculate ACK Dispersion based on the measured ACK train disp. and consider the Delayed ACK counter. This is done only once per ACKED BURST.
			if (m_delAckCount == 0)
			{
				ackDispersion = MicroSeconds(ackTrainDispersion.GetMicroSeconds()/ (ceil((double)bWnd / m_segmentSize) - 1));
			}
			else
			{
				ackDispersion = MicroSeconds(ackTrainDispersion.GetMicroSeconds()/ (ceil((double)bWnd / m_segmentSize) - m_delAckCount));
			}

			// 3- Calculate AvgRTT and Filter the high values using a weighted filter. This is done only once per ACKED BURST.

			double filter = 0;

			if (m_avgRtt < 0)
			{
				m_avgRtt = m_minRtt;
			}

			if (m_minRtt > 0)
			{
				rtt = rtt - ackTrainDispersion;
				filter = (rtt.GetMicroSeconds() - m_minAbsRtt.GetMicroSeconds())/double(rtt.GetMicroSeconds());
				filter = (filter < 0) ? 0 : filter;
				filter = (filter > 1) ? 1 : filter;
				m_avgRtt = MicroSeconds(filter * m_avgRtt.GetMicroSeconds() + m_minAbsRtt.GetMicroSeconds());
			}

			if ((bWnd == m_initialBWnd && ackDispersion.GetSeconds() != 0.0) || m_ackDispersion.GetSeconds() == 0.0)
			{
				m_ackDispersion = ackDispersion;
			}

			NS_ASSERT(m_ackDispersion.GetSeconds() != 0.0);

			// 4- Calculate the Delta RTT value,  this is done once per ACKED BURST.
			deltaRtt = m_avgRtt - m_minAbsRtt;

			if (deltaRtt.GetMilliSeconds() < 1)
			{
				deltaRtt = MilliSeconds(1);
			}
			// call Train Received function to select the burst mode (tracking/ adjustment) based on deltaRTT & Beta value.
			TrainReceived(bWnd, ackTrainDispersion, deltaRtt);
		}

		// Second ACCE FUNCTION: BWND history pop-up
		m_historyBWnd.pop_front();
		m_maxReTx = 0;
		m_ACCEProcessing = false;
		m_bytesAcked -= bWnd;

		// reset min, max RTTs
		m_minRtt = Time::Max();
		m_maxRtt = Time::Min();

		m_firstAck = Time::FromInteger(0, Time::MS);

		NS_ASSERT(m_bWnd <= m_initialBWnd);
	}
}

void
TcpWave::TrainReceived(uint32_t bWnd, Time ackTrainDispersion,Time deltaRtt)
{
	NS_LOG_FUNCTION(this << bWnd << ackTrainDispersion);
	++m_trainReceived;

	if (m_trainReceived < m_sf)
	{
		return;
	}
	if (ackTrainDispersion.GetSeconds() != 0.0 && deltaRtt.GetSeconds() != 0.0)
	{
		if (deltaRtt > m_congThresold)
		{
			RateAdjustment(bWnd, ackTrainDispersion, deltaRtt);

			m_sf = (m_avgRtt.GetMicroSeconds()/ (m_ackDispersion.GetMicroSeconds() * (m_initialBWnd / m_segmentSize)));
			m_sf = (m_sf > STAB) ? STAB : STAB;

		}
		else
		{
			RateTracking(bWnd, m_ackDispersion);
			m_sf = STAB;
		}
	}

	m_trainReceived = 0;
}

void
TcpWave::RateAdjustment(uint32_t bWnd, const Time& ackTrainDispersion,const Time& deltaRtt)
{
	NS_LOG_FUNCTION(this << ackTrainDispersion.GetMilliSeconds () << deltaRtt.GetMilliSeconds ());
	m_txmode = false;
	NS_ASSERT(deltaRtt.GetSeconds() != 0.0);
	NS_ASSERT(ackTrainDispersion.GetSeconds() != 0.0);
	uint32_t oldBWnd = m_bWnd;
	double tmp = (ackTrainDispersion.GetSeconds()/ (ackTrainDispersion.GetSeconds() + deltaRtt.GetSeconds()));
	NS_ASSERT(tmp <= 1.0);
	m_bWnd = m_bWnd * tmp;
	RoundSupBurstWnd();

	if (m_bWnd > BURST_MIN )
	{
		m_txTime = MilliSeconds((m_initialBWnd / m_segmentSize)* m_ackDispersion.GetMilliSeconds());

	}
	else
	{
		m_txTime = MilliSeconds(m_txTime.Get().GetMilliSeconds() * LAMBDA);
		m_txTime =(m_txTime.Get().GetMicroSeconds()> m_initialTxTime.GetMicroSeconds()) ? m_initialTxTime : m_txTime.Get();
		m_bWnd = BURST_MIN; // preserve minimum BURST size
		m_minAbsRtt = m_avgRtt; //Time::Max ();
		m_avgRtt = m_maxRtt;
	}

	if (m_txTime.Get().GetSeconds() == 0.0)
	{
		m_txTime = m_initialTxTime;
	}

	uint32_t diff = oldBWnd - m_bWnd;
	if (oldBWnd > m_bWnd && !m_recovery) {
		if (m_cWnd - UnAckDataCount() > diff)
		{
			m_cWnd -= diff;
		}
		else
		{
			m_cWnd = UnAckDataCount();
		}

		NS_ASSERT(m_bWnd >= m_cWnd - UnAckDataCount());
	}
	RoundTxTime();
}

void
TcpWave::RateTracking(uint32_t bWnd, Time ackDispersion)
{
	NS_LOG_FUNCTION(this);
	m_txmode = true;
	NS_ASSERT(ackDispersion.GetSeconds() != 0.0);
	uint32_t bytesToAdd = (m_initialBWnd - m_bWnd) / 2;
	m_bWnd += bytesToAdd;
	RoundSupBurstWnd();
	Time deltaT = (m_avgRtt - m_minAbsRtt);
	Time zeroT = ns3::Time::FromInteger(0, Time::MS);

	m_txTime = MicroSeconds((m_initialBWnd / m_segmentSize) * ackDispersion.GetMicroSeconds()+ 0.5* (std::max(deltaT.GetMicroSeconds(),zeroT.GetMicroSeconds())));
	m_txTime =(m_txTime.Get() > m_initialTxTime) ? m_initialTxTime : m_txTime.Get();
	RoundTxTime();
}

void
TcpWave::RoundSupBurstWnd()
{
	if (m_bWnd % m_segmentSize != 0)
	{
		uint32_t diff = m_bWnd % m_segmentSize;
		m_bWnd -= diff;
		m_bWnd += m_segmentSize;
	}

	if (m_bWnd > m_initialBWnd)
	{
		m_bWnd -= m_segmentSize;
	}

	NS_ASSERT(m_bWnd <= m_initialBWnd);
}

void
TcpWave::RoundTxTime()
{
	while (m_txTime.Get().GetSeconds() >= 2*m_initialTxTime)
	{
		m_txTime = m_initialTxTime;
		RoundSupBurstWnd();
	}
	if (m_bWnd < BURST_MIN)
	{
		m_bWnd = BURST_MIN;
	}
}
void
TcpWave::DupAck(const TcpHeader& header, uint32_t count)
{
	Time now = Simulator::Now();
	NS_LOG_FUNCTION(this << "tcpHeader " << count);

	if (m_bytesRetransmitted >= m_bWnd)
	{
		return;
	}

	if (header.GetAckNumber()<= m_rtoTxThreshold || m_recovery == true)
	{
		return;
	}

	if (count == 3)
	{
		m_bytesRetransmitted = m_segmentSize;
		DoRetransmit();
	}
}

void
TcpWave::Retransmit(void)
{

	NS_LOG_FUNCTION(this);

	// If erroneous timeout in closed/timed-wait state, just return
	if (m_state == CLOSED || m_state == TIME_WAIT) {
		return;
	}

	// If all data are received (non-closing socket and nothing to send), just return
	if (m_state <= ESTABLISHED && m_txBuffer->HeadSequence() >= m_highTxMark) {
		return;
	}

	if (m_rto.Get().GetSeconds() == 60 && m_state == FIN_WAIT_1)
	{
		TcpSocketBase::TimeWait ();
		return;
	}

	m_bWnd = m_initialBWnd;
	m_txTime = m_initialTxTime;
	m_cWnd = m_initialBWnd;
	m_bytesRetransmitted = 0;
	m_maxReTx = 0;
	m_recovery = true;
	m_historyTransmitTime.clear();
	m_avgRtt = 0 * m_avgRtt;
	m_minAbsRtt = m_avgRtt;

	//Keep track of last ever sent packet as target for RTO recovery retransmission
	m_rtoTxThreshold = m_highTxMark.Get();
	m_nextTxSequence = m_txBuffer->HeadSequence();
}

Ptr<TcpSocketBase>
TcpWave::Fork(void)
{
	return CopyObject<TcpWave>(this);
}
