/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2015 Satellite Multimedia Group (TLCSat)
 * http://www.tlcsat.uniroma2.it
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * Authors: A. Abdelsalam, M. Luglio, C. Roseti and F. Zampognaro, <{abdel.salam, roseti, zampognaro}@ing.uniroma2.it>
 *
 * *** brief TcpWave implementation *****
 * TCP Wave is a new transport protocol designed to provide a reliable transport services over the current and future networks.
 * This goal is achieved through an innovative burst-based transmission paradigm, as an alternative of the traditional TCP window based approach.
 * TCP Wave design principles have been tailored to assure an optimal performance under time-varying network topologies.
 * To follow the protocol implementation and how the burst transmission is performed, start from the function
 * TcpWave::SendPendingData.
 */

#ifndef TCPWAVE_H
#define TCPWAVE_H
#include "ns3/tcp-socket-base.h"
#include "ns3/timer.h"
#include <deque>
#include <map>
#include <iostream>
#include "tcp-tx-buffer.h"
#include "tcp-rx-buffer.h"

namespace ns3 {
typedef std::pair<SequenceNumber32, SequenceNumber32> Range32;
typedef std::pair<Range32, Time> PairOfRangeAndTime;
typedef std::map<Range32, Time> MapOfTransmitTime;

class TcpWave : public TcpSocketBase
{
public:
	static TypeId GetTypeId (void);

	TcpWave ();

protected:
	virtual int Listen (void);
	virtual int Connect (const Address &address);
	virtual void CancelAllTimers (void);
	virtual void NewAck (SequenceNumber32 const& rAck); // process each ACK received, update ACCE function with respect to ACK.
	virtual Ptr<TcpSocketBase> Fork (void);
	virtual void DupAck (const TcpHeader& t, uint32_t count); // senses for the received DupAcks, initiate the FR phase.
	virtual void Retransmit (void); // Senses for RTO expiration, trigger RTO recovery phase.
	virtual bool SendPendingData (bool withAck = false); // Called by the application to send data. It sends as much data available with respect to the available m_CWND size, as inflated by CallSendPendingData function.
	virtual uint32_t Window (void); // Initiate the TxTimer as soon there is data available from the APP. Called by SocketBase to Return the amount of unacked bytes.
	virtual uint32_t AvailableWindow (void);
	void GetHistoryBWndContent(void); // print out the contents and size of BWND history array

	//** Interacting with the ns3::TcpSocket function with Attributes get/set.*/
	virtual void     SetInitialSSThresh (uint32_t threshold);
	virtual uint32_t GetInitialSSThresh (void) const;
	virtual void     SetInitialCwnd (uint32_t cwnd);
	virtual uint32_t GetInitialCwnd (void) const;


	/** The Initial burst size (BWND0) */
	uint32_t m_initialBWnd;

	/** the Actual burst window */
	uint32_t m_bWnd;
	std::deque<uint32_t> m_historyBWnd;

	/** Actual congestion window */
	uint32_t m_cWnd;

	/** Initial TxTime (TxTime0) */
	Time m_initialTxTime;

	/** Internal transmission timer, started by Window function, once there is a data available from the APP, rescheduled by the calculated m_txTime */
	Timer m_txTimer;

	/** Actual tx timer, updated based on the calculated statistics W.R.T received ACK */
	TracedValue<Time> m_txTime;

	/** safety check bool for m_txTimer timeout */
	bool isTimerEnabled;


	virtual void RateTracking (uint32_t bWnd, Time ackDispersion); /**Tracking mode transmission */
	virtual void RateAdjustment (uint32_t bWnd, const Time& delta, const Time& deltaRtt); /**Adjustment mode transmission */

	void RoundSupBurstWnd (); /** Round-Up the Burst size W.R.T the MSS segmentsize */
	void RoundTxTime (); /** limits the TxTime value to the max of initial TxTime */

private:
	void StartTxTimer (); // initiates the TxTimer, calls CallSendPendingData to request transmission, reschedule the TxTime whenever there is a data available
	void CallSendPendingData (); // calculates the right amount of bytes to send in (cWnd), add it to the actual UnAckDataCount() to create the right m_cWnd to use.
	void InitializeCwnd (void); // set the initialBWnd in segments, as specified by socket SetInitialCwnd.

	MapOfTransmitTime::iterator GetIteratorOf (SequenceNumber32 seq); // reads the m_historyTransmitTime enteries.
	Time GetTransmitTime (SequenceNumber32 seq); // Get the actual transmission time for transmitted bytes from m_historyTransmitTime.
	void DiscardTimeUpTo (SequenceNumber32 seq); // Clears m_historyTransmitTime entries as per received ACKs.

	void ACCE (uint32_t bytesAcked, Time rtt, Time now); //Updates the transmission statistics based on the received ACKs.
	void TrainReceived (uint32_t bWnd, Time ackTrainDispersion, Time deltaRtt); // switches between Tracking / Adjustment modes

	uint32_t m_trainReceived; // ignored, currently not in use
	uint32_t m_bytesAcked;
	uint32_t m_maxReTx;
	Time m_minRtt;
	Time m_maxRtt;
	Time m_avgRtt;
	Time m_minAbsRtt; // the min RTT ever registered
	uint32_t m_sf; // ignored, currently not in use
	Time m_minTrainDispersion; // fallback reference for the un measured ackTrainDispersion
	Time m_firstAck; // time the first ack in the train is received
	SequenceNumber32 m_rtoTxThreshold; // the highest transmitted byte at the time of RTO expiration
	SequenceNumber32 m_head; // head of transmission burst
	Time m_congThresold; // Beta value, predefiend, (i.e. Config::SetDefault ("ns3::TcpWave::B", TimeValue (MilliSeconds (350))))
	Time m_ackDispersion; // Dispirtion of ACKs within the train
	uint32_t m_bytesRetransmitted; // counter for transmitted bytes
	uint32_t m_bytesTransmitted; // counter for the re-transmitted bytes
	bool m_recovery; //RTO recovery on/off switch
	bool m_txmode; // Tracking/Adjustment transmission (for debug, output analysis)
	bool m_ACCEProcessing; // ACCE stat. on/off based on the received Cum. ACK and the number of ACKED SEGMENTS.
	MapOfTransmitTime m_historyTransmitTime; // Transmitted SEQ, transmission time array
};

} // namespace ns3

#endif // TCPWave_H
